/*
 *
 * Multithreaded HTTP Download Accelarator: Aget
 * (c) 2002, Murat Balaban <murat@enderunix.org>
 *
 * See COPYING for copyright and copying restrictions
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <netdb.h>
#include <pthread.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <sys/socket.h>
#include <sys/types.h>

#include "socket.h"
#include "misc.h"


#define	HTTPPORT	80
#define FTPREQ	21	
#define UNKNOWNREQ	2

#define MAXTHREADS	10


int main(int argc, char **argv)
{
	struct request req;
	pthread_t *threads[MAXTHREADS];
	pthread_t *tid;
	thread_data *tdata;
	int i, c, error = 0, r; 
	int nthreads = 0;
	int soffset, foffset;
	FILE *fp, *tmpfp;
	extern char *optarg;
	extern int optind;
	char *rbuf, *tmpfile;
	char *fullurl;

	rbuf = (char *)calloc(BUFSIZ, sizeof(char));
	tmpfile = (char *)calloc(MAXFILESIZ, sizeof(char));

	while (!error && (c = getopt(argc,argv,"n:hv")) != -1) {
		switch(c) {
			case 'n':
				nthreads = atoi(optarg);
				break;
			case 'h':
				printf("%s\n", PROGVERSION);
				usage();
				exit(0);
				break;
			case 'v':
				puts(PROGVERSION);
				exit(0);
				break;
			default:
				error = 1;
				usage();
				exit(1);
				break;
		}
	}

	if (error) {
		usage();
		exit(1);
	}


	if (argc == 2) {		/* If only url is supplied...	*/
		fullurl = strdup(argv[1]);
		printf("Warning: number of threads is not specified, defaulting to 1\n");
		nthreads = 1;
	} else
	if (optind < argc)
		if (argc > 2)
			fullurl = strdup(argv[optind]);
		else {
			usage();
			exit(1);
		}
	else
	if (optind == argc) {
		usage();
		exit(1);
	}



	parse_url(fullurl, &req);
	
	head_req(&req);
	printf("<LOG> Downloading %s from site %s(%s:%d). Number of threads: %d\n",
			req.url, req.host, req.ip, req.port, nthreads);
	printf("<LOG> Local File: %s\n", req.file);

	for (i = 0; i < nthreads; i++) {
		soffset = calc_offset(req.clength, i, nthreads);
		foffset = calc_offset(req.clength, i + 1, nthreads);
		tid = (pthread_t *)malloc(sizeof(pthread_t));
		tdata = (thread_data *)malloc(sizeof(thread_data));
		strncpy(tdata->host, req.host, MAXHOSTSIZ);
		strncpy(tdata->url, req.url, MAXURLSIZ);
		strncpy(tdata->ip, req.ip, MAXIPSIZ);
		strncpy(tdata->file, req.file, MAXFILESIZ);
		tdata->soffset = soffset;
		tdata->foffset = (i == nthreads - 1 ? req.clength : foffset);
		tdata->port = req.port;
		tdata->fp = fp;
		tdata->tind = i;
		thread_exit_cnt++;
		pthread_create(tid, NULL, partialGetHandler, tdata);
		threads[i] = tid;
	}
	threads[i] = NULL;

	/* Later !
	pthread_mutex_lock(&thread_exit_mtx);
	while (thread_exit_cnt != 0) 
		pthread_cond_wait(&thread_exit_cnd, &thread_exit_mtx);
	pthread_mutex_unlock(&thread_exit_mtx);
	*/

	for (i = 0; threads[i] != NULL; i++) 
		pthread_join(*threads[i], NULL);

	Log("Download complete, assembling file...");

	if ((fp = fopen(req.file, "w")) == NULL) {
		perror("fopen");
		exit(1);
	}

	for (i = 0; i < nthreads; i++) {
		sprintf(tmpfile, "%s.tmp%d", req.file, i);
		if ((tmpfp = fopen(tmpfile, "r")) == NULL) {
			perror("fopen");
			exit(1);
		}
		while ((r = fread(rbuf, sizeof(char), BUFSIZ, tmpfp)) > 0) 
			fwrite(rbuf, sizeof(char), r, fp);
		fclose(tmpfp);
		unlink(tmpfile);
	}


	Log("File assembly finished, job completed successfully!");
	Log("Shutting down...");
	fclose(fp);
	return 0;
}
